// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *		   Grayscale Image Filtration
 *
 * Declaration of functions/procedures (generally grouped together
 * under the FilterIt class umbrella) that do linear and some non-linear
 * image filtration and related transfromations. Most of the operations
 * do filtration "inplace"
 *
 * The FilterIt class is largely a syntactic sugar; you can't explicitly
 * construct an object of this class, on stack or on heap. The only way
 * the class can be used is within a pattern
 *	filter(image).median(RowsAndColumns);
 * etc. Besides, this is the only way it makes sense.
 *
 * Note that class FilterIt is a friend of IMAGE, so it's privy of
 * IMAGE internals and can use them to make processing faster.
 *
 * $Id: filter.h,v 2.2 1996/07/15 19:45:11 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#endif
#ifndef _filter_h
#define _filter_h

#ifdef __GNUC__
#pragma interface
#endif

#include "image.h"
#include "std.h"

/*
 *------------------------------------------------------------------------
 * 			Convolution kernels
 *
 * Kernel coeffs are stored as rational numbers with some common denominator,
 * which can be either 1 (that is, all coeffs are ints) or power of two, or
 * some arbitrary card number. We use a family of CommonDenom classes for each
 * of those situations. Each members of the family defines how to
 * execute a division of long int by a denominator most efficiently.
 * The classes are then used in "templates" in actual filtration algorithm
 */

class CommonDenomOne
{
public:
  inline int divide(long int x)	const { return x; }
};

class over_2_up
{
  card exponent;
public:
  over_2_up(const card _exp) : exponent(_exp) {}
  inline int divide(long int x)	const { return x>>exponent; }
};

class CommonDenom
{
  card denom;
public:
  CommonDenom(const card _denom) : denom(_denom) {}
  inline int divide(long int x)	const { return x/(signed)denom; }
};

				// 1D 3-point convolution kernel
template <class Denom>
class ConvKernel3
{
  const int am1,a0,ap1;		// a[k-1], a[k], a[k+1] coeffs
  const Denom common_denominator;
public:
  ConvKernel3(const int _am1, const int _a0, const int _ap1,
	      const Denom denom) :
	am1(_am1), a0(_a0), ap1(_ap1), common_denominator(denom) {}
  inline int convolve(const GRAY_SIGNED pm1, const GRAY_SIGNED p0,
		      const GRAY_SIGNED pp1) const
  { return common_denominator.divide(pm1*am1 + p0*a0 + pp1*ap1); }
};

				// A 3x3 matrix 
struct Matrix3x3
{
  int app, ap0, apn;			// Coeffs centered around a00
  int a0p, a00, a0n;
  int anp, an0, ann;

  void set_right_center_cols(const int _ap0, const int _a00, const int _an0)
  { apn = ap0 = _ap0; a0n = a00 = _a00; ann = an0 = _an0; }

  int sum(void) const
  { return app + ap0 + apn + a0p + a00 + a0n + anp + an0 + ann; }

  			// shift in three numbers into the right column,
  			// shoving the other columns to the left
  void shift_in_next_col(const int _apn, const int _a0n, const int _ann)
  {
    app = ap0; a0p = a00; anp = an0;
    ap0 = apn; a00 = a0n; an0 = ann;
    apn =_apn; a0n =_a0n; ann =_ann;
  }
  
  			// Just shift in the right column again
  			// (useful at the right edge of an image)
  void shift_in_next_col(void)
  {
    app = ap0; a0p = a00; anp = an0;
    ap0 = apn; a00 = a0n; an0 = ann;
  }
  			// compute the inner product
  int operator * (const Matrix3x3& am) const
  { return app * am.app + ap0 * am.ap0 + apn * am.apn +
  	   a0p * am.a0p + a00 * am.a00 + a0n * am.a0n +
  	   anp * am.anp + an0 * am.an0 + ann * am.ann; }
};

				// 2D 3-point convolution kernel
template <class Denom>
class ConvKernel3x3
{
  const Matrix3x3 kernel;
  const Denom common_denominator;
public:
  ConvKernel3x3(const Matrix3x3& kernel_matrix,const Denom denom) :
	kernel(kernel_matrix), common_denominator(denom) {}
  inline int convolve(const Matrix3x3& pixels) const
  { return common_denominator.divide(pixels * kernel); }
};


				// Convolution kernels
					// all coeffs integers
inline ConvKernel3<CommonDenomOne> conv_kernel(
	const int am1, const int a0, const int ap1)
{ return ConvKernel3<CommonDenomOne>(am1,a0,ap1,CommonDenomOne()); }

inline ConvKernel3<over_2_up>
conv_kernel(const int am1, const int a0, const int ap1, over_2_up denom)
{ return ConvKernel3<over_2_up>(am1,a0,ap1,denom); }

inline ConvKernel3<CommonDenom>
conv_kernel(const int am1, const int a0, const int ap1, CommonDenom denom)
{ return ConvKernel3<CommonDenom>(am1,a0,ap1,denom); }

inline ConvKernel3x3<CommonDenomOne>
conv_kernel(const Matrix3x3& kernel_matrix)
{ return ConvKernel3x3<CommonDenomOne>(kernel_matrix,CommonDenomOne()); }

inline ConvKernel3x3<over_2_up>
conv_kernel(const Matrix3x3& kernel_matrix, over_2_up denom)
{ return ConvKernel3x3<over_2_up>(kernel_matrix,denom); }

inline ConvKernel3x3<CommonDenom>
conv_kernel(const Matrix3x3& kernel_matrix, CommonDenom denom)
{ return ConvKernel3x3<CommonDenom>(kernel_matrix,denom); }


/*
 *------------------------------------------------------------------------
 * 				Look-up tables
 * to specify a mapping operation GRAY -> GRAY
 */

class LookupT
{
  friend class FilterIt;

  GRAY min_val, max_val;		// in the GRAY domain of the mapping
  GRAY * table;				// mapping itself
  card no_entries;			// in the table

					// init the LookupT but don't fill in
					// the table
  void allocate(const GRAY _min_val, const GRAY _max_val);

  enum { Preallocated_no = 256 };	// Allocate ahead that many entries
  GRAY preallocated [Preallocated_no];

public:
				// What to do with pixels that outside
				// [min_val,max_val] when doing mapping
  enum FringeHandling { CoerceFringes, LeaveFringes };

				// Just to make constructing some common
				// look-up tables sweeter...
  struct Identical { const card depth; Identical(const card _d) : depth(_d){}};
  struct MapTo {
    const GRAY from, to;
    MapTo(const GRAY _from, const GRAY _to) : from(_from), to(_to) {}
  };
  struct DownGRAY {
    const card depth_from, depth_to;
    DownGRAY(const card _f, const card _t) : depth_from(_f), depth_to(_t) {}
  };


				// Identical pixel mapping for an image of
				// specified depth
  LookupT(const Identical id);

				// Build a table (with one entry) to map
				// just one particular pixel
  LookupT(const MapTo map_one);

				// Build a table to reduce image depth from
				// one value to another (that is, build a
				// stripy table)
  LookupT(const DownGRAY downgray);

  LookupT(const LookupT& another);	// standard assignment stuff
  LookupT& operator= (const LookupT& another);
  ~LookupT(void);

				// Getting hold of and/or modifying
				// specific mappings
  GRAY& operator[] (const GRAY val);
  GRAY  operator[] (const GRAY val) const;

				// Convolve two Lookup tables, that is,
				// translate pixels of this according to
				// another table
				// (Lookup table composition)
  LookupT& apply(const LookupT& another,
		 const FringeHandling fringe_handling);

  GRAY q_min(void) const	{ return min_val; }
  GRAY q_max(void) const	{ return max_val; }
};

inline GRAY& LookupT::operator[] (const GRAY val)
{
  if( val < min_val || val > max_val )
    _error("The pixel value %d to look up is out of range [%d,%d]",
	   val,min_val,max_val);
  return table[val-min_val];
}

inline GRAY  LookupT::operator[] (const GRAY val) const
{ return (*(LookupT *)this)[val]; }

/*
 *------------------------------------------------------------------------
 * 		    The umbrella class FilterIt itself
 *
 */

class FilterIt
{
  IMAGE& image;				// Image under filtration

  void operator=(const FilterIt&);	// is not implemented, ergo, is
					// not allowed

					// A private constructor, to make
					// sure the object can't be constructed
					// and left, while image would disappear
					// Use a filter() method to do
					// the actual construction
  FilterIt(IMAGE& _image) : image(_image) { image.is_valid(); }
  
  FilterIt(const FilterIt& ft);		// Is private and not used: just
  					// to discourage any unintended use of
  					// this class

  IMAGE& median_3_row(void);		// 3-point row-wise filtration
  IMAGE& median_3_col(void);		// 3-point col-wise filtration
  IMAGE& median_5_row(void);		// 5-point row-wise filtration
  IMAGE& median_5_col(void);		// 5-point col-wise filtration

public:
  
				// How to apply filters
  enum Direction { Columns, Rows, RowsAndColumns };

					// FilterIt(IMAGE& image) would be
					// implicitly called
  friend inline FilterIt filter(IMAGE& image) { return image; }

  IMAGE& median(const Direction how, const int win_size = 3);

				// Convolution methods
				// Templates don't work inside the class 8:(
				// ("method templates" would do, but they
				// aren't supported yet)
#define CONV_DECL(Kernel,Denom)				\
  IMAGE& conv_row(const Kernel<Denom>& kernel);		\
  IMAGE& conv_col(const Kernel<Denom>& kernel);		\
  IMAGE& conv(const Kernel<Denom>& kernel,		\
	      const Direction how = RowsAndColumns);	\

  CONV_DECL(ConvKernel3,CommonDenomOne)
  CONV_DECL(ConvKernel3,over_2_up)
  CONV_DECL(ConvKernel3,CommonDenom)

#undef CONV_DECL

				// Convolution with non-separable kernels
  IMAGE& conv(const ConvKernel3x3<CommonDenomOne>& kernel);
  IMAGE& conv(const ConvKernel3x3<over_2_up>& kernel);
  IMAGE& conv(const ConvKernel3x3<CommonDenom>& kernel);
  
  				// Non-determenistic averaging: some
  				// kind of simulated annealing
  IMAGE& nondet_average(const int temperature);

				// Translate image pixels according to
				// the look-up table
  IMAGE& translate(const LookupT& lookupt,
		   const LookupT::FringeHandling fringe_handling);
};

#endif
