// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *   			Generating Fractal Maps
 *
 * The inspiration for the original algorithm was drawn from the following
 * explanation described in the "documentation" to program MARS 
 * (Tim Clarke, tjc1005@hermes.cam.ac.uk, posted in some "games"
 * group)
 *
 * "This is a recursive subdivision, or plasma, fractal. You start of
 * with a random height at (0,0) and therefore also at (256,0), (0,256),
 * (256,256).  Call a routine that takes as input the size and position
 * of a square, in the first case the entire map.
 * This routine get the heights from the corners of the square it gets
 * given.  Across each edge (if the map has not been written to at the
 * point halfway along that edge), it takes the average of the heights of
 * the 2 corners on that edge, applies some noise proportional to the
 * length of the edge, and writes the result into the map at a position
 * halfway along the edge. The center of the square is the average of the
 * four corners+noise.
 * The routine then calls itself recursively, splitting each square into
 * four quadrants, calling itself for each quadrant until the length of
 * the side is 2 pixels.
 * This is probably old-hat to many people, but the map is made more
 * realistic by blurring [applying some filter to it]
 * The colors are done so that the sun is on the horizon to the East:
 *     Color=A*[ w(u+1,v)-w(u,v) ]+B
 * with A and B chosen so that the full range of the palette is used.
 * The sky is a similar fractal but without the colour transformation"
 * <EOQ>
 *
 * A high spirit (and only the general spirit) of the algorithm was
 * implemented in the old FractalNPMap generator. Note, that *unlike*
 * the technique described above, the implementation is *iterative* rather
 * than recursive. For another thing, our map is *NOT* periodic. BTW, 
 * it lets one use seeds in every corner of a map (which would create "biased"
 * clouds).
 *
 * Further inspiration was taken from Matt's Fractals (Matson Dawson,
 * Matt, University of Canterbury, New Zealand, Misc1666@csc.canterbury.ac.nz)
 * That led to a new version of the algorithm, FractalMap. Again, only
 * the inspiration was taken from that code: the implementation is completely
 * different, and _highly_ optimized and efficient.
 *
 * A single step of the old algorithm performed an averaging followed by
 * a random displacement. In contrast, the new algorithm does the
 * random displacement first, and averaging later. This tends to produce 
 * smoother (and thus, more appealing) clouds.
 *
 * In the old algorithm, a map's pixel value is computed as average + noise,
 * the noise being scaled up to the current dim. So when 2*scale > max pixel
 * value, we've got a problem. To get around that, the scale factor of the
 * noise is chosen not to exceed max_pixel_value/2. In the new implementation,
 * this problem is handled with more finesse: the seed is set to 
 * max_pixel_value/2, noise_scale is set to a half of that, 
 * resulting pixel values are always clipped, and noise_scale is 
 * reduced independently of the averaging scale.
 *
 * In the old algorithm, since the map was not-periodical, if a pixel sticks
 * out of the map, we force it within: say, when we need to operat
 * map(2^order,j), we actually dealing with map((2^order)-1,j), etc.
 * The new algorithm makes self-tiling (periodical) maps; That is,
 * 	map(i+k*dim,j+l*dim) = map(i,j) for any k,l
 * In constructing a map, the algorithm never uses indices greater than dim;
 * and indices = dim are tacitly mapped to 0.
 *
 * Note, the initial seed is the same as the initial noise factor (which
 * is usually the middle of the pixel range). Thus we utilize the pixel
 * dynamic range as much as possible. If one wishes to use a different
 * seed, it's is always possible to darken/brighten the image after it
 * was constructed (by doing a LookUp table substitution).
 *
 * $Id: fractal_map.cc,v 2.1 1996/07/09 17:05:24 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation "fractal_map.h"
#endif

#include "fractal_map.h"
#include "minmax.h"
#include "std.h"

				// Get some noise with a dispersion 'scale'
				// (which is assumed to be a power of 2)
				// and average 0
				// Say, if scale=2 return either 0 or 1
				// if scale=128, return a random number
				// within [-64,63]
				// Note, that usually a couple of low-order
				// bits of rand() are "less random";
				// therefore, we shift them out.
inline int FractalNPMap::get_noise(const card scale) const
{
  return ( (rand()>>3) & (scale-1) ) - scale/2;
}

inline int FractalMap::get_noise(const card scale) const
{
  return ( (rand()>>3) & (scale-1) ) - scale/2;
}

FractalMap::FractalMap
	(const card order, const card smoothness, const card starting_scale_log,
	 const card bits_per_pixel)
	: LazyImage(1<<order,1<<order,bits_per_pixel),
	  noise_starting_factor(1<<(bits_per_pixel-1)),
	  starting_scale(1<<(starting_scale_log == 0 ? min(order,(card)6) : 
	  				               starting_scale_log)),
	  noise_cutoff_scale(1<<smoothness)
{
  if( order < 2 || order > 14 )
    _error("Order %d of the desired fractal map is unusual, "
	   "I prefer something within [2,14]",order);
  if( starting_scale < 2 || starting_scale > nrows )
    _error("Starting scale %d must be within [2,2^order=%d]",
    	   starting_scale,nrows);
  if( noise_cutoff_scale >= starting_scale )
    _error("2^smoothness %d exceeds starting_scale %d: the resulting image "
    	   "would be completely uniform",noise_cutoff_scale,starting_scale);
}

//------------------------------------------------------------------------
//		Here where the map is really constructed

	// returns x % dim, for a particular case when dim is an
	// even multiple of x, and x can't actually exceed dim
static inline card fast_mod(const card x, const card dim)
{ return x >= dim ? 0 : x; }


void FractalMap::fill_in(IMAGE& im) const
{
  const int max_pixel_val = (1<<im.q_depth())-1;
  const card dim = im.q_nrows();
  register card i,j;
  
  int noise_factor = noise_starting_factor;
  for(i=0; i < dim; i += starting_scale)	// Seed the map: seed = noise_factor
    for(j=0; j < dim; j += starting_scale)	// to fully utilize the dynamic pixel
       im(i,j) = noise_factor;			// range

  for(register card scale=starting_scale; scale>1; scale >>= 1)
  {
    const card scale_half = scale>>1;

    				// Change pixels at 'scale' range at random
    if( scale > noise_cutoff_scale )
    {
     for(i=0; i<dim; i += scale)
      for(j=0; j<dim; j += scale)
      {
      	GRAY_SIGNED& pixel = (GRAY_SIGNED&)im(i,j);
      	pixel += get_noise(noise_factor);
      	if( pixel < 0 )
      	 pixel = 0;
        else if( pixel > max_pixel_val )
         pixel = max_pixel_val;
      }
      noise_factor = max(1,noise_factor>>1);
    }
    				// Compute pixels at the "half-step"
    				// by linear interpolation (averaging
    				// its neighbors)
    for(i=0; i<dim; i += scale)
    {
      const int i_up = fast_mod(i+scale,dim);
      const int i_hp = i+scale_half;
      int corner00 = im(i,0);			// caching
      int corner10 = im(i_up,0);
      for(j=0; j<dim; j += scale)
      {
	const int j_up = fast_mod(j+scale,dim);
	const int j_hp = j+scale_half;
	const int corner01 = im(i,j_up);
	const int corner11 = im(i_up,j_up);
	im(i_hp,j) = (corner00 + corner10)>>1;
	im(i,j_hp) = (corner00 + corner01)>>1;
	im(i_hp,j_hp) = (corner00 + corner10 + corner01 + corner11)>>2;
				// Note, corner01 for col j becomes
	corner00 = corner01;	// corner00 for the next j
	corner10 = corner11;
      }
    }
  }
}


//------------------------------------------------------------------------
//		The old generation of the FractalMap
// It constructs a non-tiling map

FractalNPMap::FractalNPMap
	(const card order, const Seeds& _seeds, const card bits_per_pixel)
	: LazyImage(1<<order,1<<order,bits_per_pixel),
	  seeds(_seeds)
{
  if( order < 2 || order > 14 )
    _error("Order %d of the desired fractal map is unusual, "
	   "I prefer something within [2,14]",order);
}


void FractalNPMap::fill_in(IMAGE& im) const
{
  const card largest_noise_scale = 1<<(im.q_depth()-1);
  const card pixel_mask = (1<<im.q_depth())-1;
  const card dim = im.q_nrows();

  im(0,0) = seeds.s00;				// Seed the map
  im(dim-1,0) = seeds.s10;
  im(0,dim-1) = seeds.s01;
  im(dim-1,dim-1) = seeds.s11;

  for(register card scale=dim; scale>1; scale >>= 1)
  {
    const card scale_half = scale>>1;
    const card noise_scale = min(scale,largest_noise_scale);
    register card i,j;
    for(i=0; i<dim; i += scale)
    {
      int i_up = min(i+scale,dim-1);
      int i_hp = i+scale_half;
      int corner00 = im(i,0);			// caching
      int corner10 = im(i_up,0);
      int mid30 = im(i_hp,0) = 
	(((corner00+corner10)>>1) + get_noise(noise_scale)) & pixel_mask;
      for(j=0; j<dim; j += scale)
      {
	int j_up = min(j+scale,dim-1);
	int j_hp = j+scale_half;
	const int corner01 = im(i,j_up);
	const int corner11 = im(i_up,j_up);
	const int mid03 = i == 0 ? 
			  ( ((corner00+corner01)>>1) + 
			    get_noise(noise_scale) ) & pixel_mask : im(i,j_hp);
	const int mid31 = ( ((corner01+corner11)>>1) + 
			    get_noise(noise_scale) ) & pixel_mask;
	const int mid13 = ( ((corner10+corner11)>>1) + 
			    get_noise(noise_scale) ) & pixel_mask;
	const int mid33 = ( ((mid30+mid03+mid31+mid13)>>2) + 
			    get_noise(noise_scale) ) & pixel_mask;
	im(i_hp,j_up) = mid31;
	if( i == 0 )
	  im(i,j_hp) = mid03;
	im(i_up,j_hp) = mid13;
	im(i_hp,j_hp) = mid33;
				// Note, corner01 for col j becomes
	corner00 = corner01;	// corner00 for the next j
	corner10 = corner11;
	mid30    = mid31;
      }
    }
  }
}


