// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *		   	 Grayscale Image
 *   		     Generating Fractal Maps
 * 
 * Declaration of a class of recursive subdivision, or plasma, fractals,
 * that generate self-similar images (with a scaled noise).
 *
 * The present file declares two generators. The old one, FractalNPMap,
 * generates non-periodical maps, that is, maps that don't tile. The
 * new generator, FractalMap's, maps do tile, that is, if you duplicate
 * the image and put it beside the original map, the two pictures
 * should blend seamlessly. The generators also differ in the algorithm:
 * the new FractalMap is faster and makes smoother images (which
 * crispness can be varied).
 *
 * A FractalMap is specified by its order, smoothness, log(starting_scale),
 * pixel depth, and a random-number generating algorithm. The order is the
 * only necessary parameter: it specifies a square map (image) of a dimension
 * 2^order. The order should be at least two (and not too large). If smoothness
 * is 0, the resulting map is going to be very rough (rugged within a pixel 
 * range); smoothness = order makes an absolutely smooth map (that is,
 * completely uniform). The starting scale (specified by its log) relates to
 * the initial distribution of the noise. The starting_scale_log must be
 * within [1,order]; it's probably best to leave it default (setting
 * the starting_scale_log to 0 means letting the program to pick a suitable
 * value). 
 *
 * A FractalNPMap is specified by its order, seeds, pixel depth, and
 * a random-number generating algorithm. Again, the order is the only
 * necessary parameter. There are reasonable defaults for the other parameters.
 * Seeds are used to specify values  at the four corners of the map, map(0,0),
 * map(2^order,0), map(0,2^order) and map(2^order,2^order). Note, strictly
 * speaking, the seeds except map(0,0) are actually outside of the image
 * itself. The map isn't periodical, btw. Map's points between the seeds
 * are filled in as an average of the corresponding boundary points
 * plus some random scaled noise. See "fractal_map.cc" for more
 * details about the implementation.
 *
 *
 * Usage examples:
 *	IMAGE map1 = FractalMap(9);	// Use all the defaults
 *	IMAGE map2 = FractalMap(9,1);
 *	class MyMap : public FractalMap
 *	{
 *	  public:
 *	  int get_noise(const card scale) const { return your_random_number; }
 *	  MyMap(const card order) : FractalMap(order) {}
 *	}
 *	IMAGE map3 = MyMap(10);		// Play with your own random number gen
 *
 * One may also want to blur/sharpen the fractal map afterwards (to make
 * it more appealing) by applying a corresponding filter, say, [1 1 1]
 *
 * Note, FractalMap is a LazyImage thing, that is, in
 *	IMAGE map1 = FractalMap(9);
 * the map1 is constructed "inplace", without *any* image copies/clones.
 *
 * $Id: fractal_map.h,v 2.1 1996/07/09 16:52:48 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#endif
#ifndef _fractal_map_h
#define _fractal_map_h

#ifdef __GNUC__
#pragma interface
#endif

#include "image.h"

class FractalMap : public LazyImage
{
  const int noise_starting_factor;
  const card starting_scale;
  const card noise_cutoff_scale;
  void fill_in(IMAGE& im) const;

public:

  FractalMap(const card order, const card smoothness = 2,
  	     const card starting_scale_log = 0,
	     const card bits_per_pixel = 8);

				// Get some noise with a dispersion 'scale'
				// (which is assumed to be a power of 2)
				// and average 0
				// Say, if scale=2 return either 0 or 1
				// if scale=128, return a random number
				// within [-64,63]
  virtual int get_noise(const card scale) const;
};

		// Non-periodic (non-tiling) map
class FractalNPMap : public LazyImage
{
public:
  class Seeds
  {
  public:
    const GRAY s00, s10, s11, s01;

    Seeds(const GRAY seed) 
      : s00(seed), s10(seed), s11(seed), s01(seed) {}
    Seeds(const GRAY ul, const GRAY ll, const GRAY ur, const GRAY lr)
      : s00(ul), s10(ll), s11(lr), s01(ur) {}
  };
  FractalNPMap(const card order, const Seeds& _seeds = Seeds(128),
	     const card bits_per_pixel=8);

				// Get some noise with a dispersion 'scale'
				// (which is assumed to be a power of 2)
				// and average 0
				// Say, if scale=2 return either 0 or 1
				// if scale=128, return a random number
				// within [-64,63]
  virtual int get_noise(const card scale) const;
private:
  Seeds seeds;
  void fill_in(IMAGE& im) const;
};

#endif


