// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   Grayscale Image
 *
 *			Non-deterministic averaging
 *
 * It looks like a regular trivial averaging/smoothing of a picture:
 * for each pixel, consider its 3x3 neighborhood (including the pixel itself),
 * and replace the pixel's value with av_value, an average brightness of this
 * 9-pixel neighborhood. The punch line is that this replacement is
 * given to a chance: we pick a random number r uniformly distributed within
 * [0,1), and if
 * 	r < exp(-abs(pixel-av_value)/temp)
 * the pixel is spared. That is, the farther a pixel sticks out from its
 * neighborhood, the higher the probability it would be shoved back in.
 * Parameter 'temp' specifies the tolerance for outstanding pixels:
 * the averaging occurs always if temp = 0, and never if temp = Infinity.
 * This is similar to a "simulated annealing", a Monte Carlo simulation
 * technique, with temp being the temperature.
 *
 * In practical terms, it's too much of a hassle dealing with exp() function,
 * so we replace exp(-x) with 1/(1+x), which has the same asymptotic behavior
 * for x >= 0. The substitute doesn't decay as fast as the real exponential
 * function though, thus more sticking out pixels would be spared than
 * Gibbs would've allowed. But we can live with that. Thus a pixel is spared
 * if
 *	r < temp/(temp + abs(pixel-av_value))
 * Let S be a normalizing factor, preferably an exact power of 2; say, 256.
 * Then an integer R = floor(r*S) is a random number uniformly distributed
 * within [0,S-1]. The condition above could be rewritten then as
 *	R*(temp + abs(pixel-av_value)) < temp*S
 * Thus temp can be integer too, and all the computation can be carried out
 * with the integer arithmetics only. Again, temp=0 makes the condition always
 * false, that is, every pixel would be replaced by the average of its
 * neighborhood. As abs(pixel-av_value) is limited to a range of
 * [0,max_pixel_value], setting temp = 10*max_pixel_value would
 * then spare even the most outstanding pixel in 90% of the cases.
 *
 * The edges of the image present a problem, as always: a 3x3 neighborhood
 * of edge/corner pixels has obviously fewer than 9 pixels. However,
 * we "extrapolate" the missing pixels according to a simple rule that
 *	pixel[-1,i] = pixel[0,i], etc.
 * That is, it is always assumed that
 *	pixel[i,j] = pixel[i%dim,j%dim] for _any_ i,j
 *
 * $Id: nondet_av.cc,v 1.1 1996/07/12 19:05:31 oleg Exp oleg $
 *
 ************************************************************************
 */

#include "filter.h"
#include <alloca.h>

IMAGE& FilterIt::nondet_average(const int temperature)
{

  		// The neighborhood of the current pixel
  class Neighborhood : Matrix3x3
  {
    enum { S_log = 8 };	// log2(Scale factor S)
    
    			// Get a random number R uniformly distributed
    			// within [0,2^S_log-1]
    			// Since the lower bits of rand() are less random,
    			// we shift them out
    int get_R(void) const
    	{ return ( rand()>>3 ) & ((1<<S_log)-1); }
  
    const int temperature;
    const int temperature_scaled;	// That is, temperature*S

  public:
  			// Populate the neighborhood starting at the
  			// left edge of the image
  			// The previous column (which does not exists
  			// actually) is "extrapolated" from the current one
  			// Note, don't forget to shift() right after the
  			// construction of the Neighborhood to propertly set
  			// all its three columns
    Neighborhood(const int _p0, const int _00, const int _n0,
    		 const int _temperature)
    	: temperature(_temperature), 
    	  temperature_scaled(_temperature<<S_log)
    	  { set_right_center_cols(_p0,_00,_n0); }
    	  
    	  		// Shift in a new column into the neighborhood,
    	  		// which becomes the rightmost column, shifting the
    	  		// others to the left
    void shift(const int _pn, const int _0n, const int _nn)
    	{ shift_in_next_col(_pn,_0n,_nn); }
    			// The same as above, only there is no next column
    			// beyond the right edge of the image. Thus
    			// we shift in the edge column once again
    void shift_end_of_row(void) { shift_in_next_col(); }
    
    		// Change the value of the curr_pixel non-deterministically
    void average(GRAY_SIGNED& curr_pixel)
    {
      const int av_val = sum()/9;
      if( get_R() * (temperature + abs(curr_pixel - av_val)) >=
          temperature_scaled )
        curr_pixel = av_val;
     }
  };
  
  if( image.ncols < 2 || image.nrows < 2)
    image.info(),
    _error("we need at least 3 rows and columns for the non-det averaging");

  register GRAY_SIGNED * pp = (GRAY_SIGNED *)image.pixels;
  GRAY_SIGNED * const prev_row = 
  	(GRAY_SIGNED *)alloca(image.ncols*sizeof(GRAY_SIGNED));
  const GRAY_SIGNED * row_end = (GRAY_SIGNED *)image.pixels + image.ncols;
  const GRAY_SIGNED * next_row = (GRAY_SIGNED *)image.pixels + image.ncols;
  const GRAY_SIGNED * const image_end = (GRAY_SIGNED *)image.pixels + image.npixels;

  {				// Handle the 1st row
    Neighborhood hood(*pp,*pp,*next_row,temperature);
    for(register GRAY_SIGNED * prp = prev_row; pp < row_end; )
    {
      GRAY_SIGNED& curr_pixel = *pp++;
      *prp++ = curr_pixel;			// save it
      if( pp == row_end )
        hood.shift_end_of_row(), ++next_row;
      else
        hood.shift(*pp,*pp,*++next_row);
      hood.average(curr_pixel);
    }
  }

  for(row_end += image.ncols; row_end <= image_end; row_end += image.ncols)
  {
    if( next_row >= image_end )		// Assume the row after the image
      next_row = pp;			// is the same as the last row

    Neighborhood hood(*prev_row,*pp,*next_row,temperature);
    for(register GRAY_SIGNED *  prp = prev_row; pp< row_end; )
    {
      GRAY_SIGNED& curr_pixel = *pp++;
      *prp++ = curr_pixel;	// curr_pixel will be needed for the next row
      if( pp == row_end )
        hood.shift_end_of_row(), ++next_row;
      else
        hood.shift(*prp,*pp,*++next_row);
      hood.average(curr_pixel);
    }
  }

  assert( pp == image_end && row_end == next_row + image.ncols );
  return image;
}
