// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   Grayscale Images
 *
 *	     Read an image from a file in the TIFF format
 *
 * The present TIFF reader reads a Class G TIFF file (for gray-scale
 * images), See Appendix G of the TIFF specification. 
 *
 * The program really needs the following tags
 *	ImageWidth
 *	ImageLength
 *	RowsPerStrip
 *	StripOffsets
 * and can't work without them. Yet, if RowsPerStrip is left out, we
 * assume there is only one strip in the image
 *
 * In addition, class G TIFF files are required to have the following
 * tags with the following values
 *	SamplesPerPixel = 1
 *	BitsPerSample   = 4 or 8
 *	Compression	= 1 or 5(LZW)
 *	PhotometricInterpretation = 0 or 1
 * So, if these tags are specified, the program makes sure that their values
 * are those that are supposed to be for class G TIFFs.
 *
 * At present, no compression is supported.
 *
 * $Id: read_tiff.cc,v 2.1 1996/03/18 19:58:25 oleg Exp oleg $
 *
 ************************************************************************
 */

#include "image.h"
#include "endian_io.h"
#include "tiff.h"
#include "minmax.h"
#include <iostream.h>

				// Read a TIFF header from the file
				// and check it
TIFFHeader::TIFFHeader(EndianIn& file)
{
  magic = file.read_short("Reading a TIFF file magic byte");
  if( magic == TIFF_BIGENDIAN )
    file.set_bigendian();
  else if( magic == TIFF_LITTLEENDIAN )
    file.set_littlendian();
  else
    _error("Unknown magic word %x",magic);
  version = file.read_short("Reading TIFF file version");
  diroffset = file.read_long("Reading the TIFF dir offset");
}

/*
 *------------------------------------------------------------------------
 *	Implementing input methods of the TIFF Directory
 *			and its entries
 */

				// Load up the item from the file: finish
				// construction of the object
void TIFFFileItem::load(EndianIn& file)
{
  tag   = file.read_short("Reading dir entry");
  type  = file.read_short("Reading dir entry");
  count = file.read_long("Reading dir entry");
  val_offset = file.read_long("Reading dir entry");
}

				// Print the contents of an item w/o frills
void TIFFDirEntry::print(void) const
{
  message("tag %d, %d ",tag,count);
  const char * type_names [] =
  { "UNKNOWNs (0)", "bytes", "characters", "short ints", "long ints",
    "rational numbers", "signed bytes", "some bytes",
    "signed short ints", "signed long ints", "signed rational numbers",
    "IEEE 4-byte floats", "IEEE 8-byte floats" };

  if( type < sizeof(type_names)/sizeof(type_names[0]) )
    message(type_names[type]);
  else
    message("unknown (%d)",type);
  message(" value %d (0x%x)\n",val_offset,val_offset);
}


TIFFFileDirectory::TIFFFileDirectory(EndianIn& file)
  : tiff_file(file)
{
  TIFFHeader header(tiff_file);
  assert( tiff_file.seekg(header.directory_offset()).good() );
  no_entries = tiff_file.read_short("no of entries");
  entries = new TIFFFileItem[no_entries];

  for(register card i=0; i<no_entries; i++)
    entries[i].load(file);

  const int offset_to_next_dir = tiff_file.read_long("reading dir");
  if( offset_to_next_dir != 0 )
    message("The TIFF file contains several images, only the first one "
	    "is going to be considered");
}

				// Destroy all entries in the directory
TIFFFileDirectory::~TIFFFileDirectory(void)
{
  assert( entries != 0 );
  delete [] entries;
}

				// Print the contents of the directory
void TIFFFileDirectory::print(void) const
{
  message("\nThere are %d entries in the TIFF file directory\nThey are\n",
	  no_entries);
  for(register card i=0;i<no_entries; i++)
    entries[i].print();
}

				// Look up an item in the directory by its
				// tag. Return 0 if not found
				// Note, we take advantage of the fact
				// that all items in the directory are
				// in the ascending order of their tags
TIFFFileItem * TIFFFileDirectory::look_up(const unsigned short tag) const
{
  card lo = 0;				// Using a binary search
  card hi = no_entries;
  while( hi > lo )
  {
    int middle = (hi+lo)/2;
    if( entries[middle].tag == tag )
      return &entries[middle];
    else if( entries[middle].tag > tag )
      hi = middle;
    else
      lo = middle+1;
  }
  return 0;				// Means not found
}


				// Look up a scalar value in the directory
unsigned long int TIFFFileDirectory::look_up(const unsigned short tag,
			    const unsigned long default_value) const
{
  TIFFFileItem * entryp = look_up(tag);
  if( entryp == 0 )
    return default_value;

  assert( entryp->tag == tag);
  assert( entryp->count == 1 );		// Check that the value is scalar
					// Note, if the value is shorter
					// than long, it's LEFT justified
  switch( entryp->type )
  {
    case TIFFDirEntry::BYTE:
    case TIFFDirEntry::ASCII:
         return (entryp->val_offset) >> 24;

    case TIFFDirEntry::SHORT:
         return (entryp->val_offset) >> 16;

    case TIFFDirEntry::LONG:
         return (entryp->val_offset);

    case TIFFDirEntry::RATIONAL:
  	 _error("Rational item is *not* a scalar one");

    default:
	 _error("Invalid type %d",entryp->type);
  }
}

				// Look up a string value in the directory
				// and return a dynamically allocated str
				// Return 0 if not found
char * TIFFFileDirectory::read_str(const unsigned short tag) const
{
  TIFFFileItem * entryp = look_up(tag);
  if( entryp == 0 )
    return 0;
  assert( entryp->tag == tag);
  assert( entryp->type == TIFFDirEntry::ASCII );
  
  int size = entryp->count;
  if( size <= 0 )
    return 0;

  char * str = new char[size];

  assert( tiff_file.seekg(entryp->val_offset).good() );

  tiff_file.read(str,size);
  if( str[size-1] != '\0' )		// Make sure the str is terminated
    str[size-1] = '\0';			// properly     

  return str;
}


				// Analyze the body of the TIFF file: strips
				// Find out where the strips are and what
				// are their dimensions. Strips themselves
				// will be loaded later.
TIFFFileStrips::TIFFFileStrips(EndianIn& file)
	: TIFFFileDirectory(file)
{
  no_cols = look_up(TIFFTAG_IMAGEWIDTH,0);
  if( no_cols == 0 )
    _error("ImageWidth tag missing or improperly specified as 0");
  
  no_rows = look_up(TIFFTAG_IMAGELENGTH,0);
  if( no_rows == 0 )
    _error("ImageLength tag missing or improperly specified as 0");

  rows_per_strip = look_up(TIFFTAG_ROWSPERSTRIP,no_rows);

  TIFFDirEntry * entryp = look_up(TIFFTAG_STRIPOFFSETS);
  assure( entryp != 0, "STRIPOFFSETS tag must be present in the TIFF file");

  no_strips = entryp->count;
  assert( no_strips > 0 );
  strip_offsets = new unsigned long[no_strips];

  if( no_strips == 1 )			// The value is coded within the
  {					// dir entry itself
    if( entryp->type == TIFFDirEntry::SHORT )
      strip_offsets[0] = (entryp->val_offset) >> 16;
    else if( entryp->type == TIFFDirEntry::LONG )
      strip_offsets[0] = entryp->val_offset;
    else
      _error("StripOffsets must contain either LONG or SHORT values");
    return;
  }

					// Otherwise we got to read them,
					// I mean, the strip offsets
  assert( tiff_file.seekg(entryp->val_offset).good() );
  
  if( entryp->type == TIFFDirEntry::SHORT )
    for(register card i=0; i<no_strips; i++)
      strip_offsets[i] = tiff_file.read_short("Reading StripOffsets");
  else if( entryp->type == TIFFDirEntry::LONG )
    for(register card i=0; i<no_strips; i++)
      strip_offsets[i] = tiff_file.read_long("Reading StripOffsets");
  else
    _error("StripOffsets must contain either LONG or SHORT values");
}

TIFFFileStrips::~TIFFFileStrips(void)
{
  assert( strip_offsets != 0 );
  delete [] strip_offsets;
}

				// set to read strip #i
void TIFFFileStrips::set_strip(const card i)
{
  assert( i < no_strips );
  assert( tiff_file.seekg(strip_offsets[i]).good() );
}

				// Read in the strips into an image
void TIFFFileStrips::assign_to(IMAGE& image)
{
  Image_iostream to_image(image);
  assert( image.q_nrows() == no_rows && image.q_ncols() == no_cols );
  const unsigned long image_size = no_rows * no_cols;
  const unsigned long strip_size = rows_per_strip * no_cols;
  register unsigned long strip, curr_pos;
  for(strip=0,curr_pos=0; strip<no_strips; strip++)
  {
    assert( tiff_file.seekg(strip_offsets[strip]).good() );

    const unsigned long curr_pos_end = min(curr_pos+strip_size,image_size);
    for(; curr_pos < curr_pos_end; curr_pos++)
      to_image.get() = tiff_file.read_byte("reading a strip");
  }
  assert( to_image.eof() );
}

/*
 *========================================================================
 *     		Root module - actual IMAGE constructor
 */

void IMAGE::read_tiff(EndianIn& file, const bool verbose)
{
  message("Reading the TIFF file\n");

  TIFFFileStrips directory(file);
  
  if( verbose )
    directory.print();

					// Checking if we can handle this
  assert( directory.look_up(TIFFTAG_SAMPLESPERPIXEL,1) == 1 );
  assert( directory.look_up(TIFFTAG_BITSPERSAMPLE,1) == 8 );
  assert( directory.look_up(TIFFTAG_COMPRESSION,COMPRESSION_NONE) == 
	 COMPRESSION_NONE );

  const int photometry = directory.look_up(TIFFTAG_PHOTOMETRIC,
				     PHOTOMETRIC_MINISBLACK);
  assert( photometry == PHOTOMETRIC_MINISBLACK ||
	  photometry == PHOTOMETRIC_MINISWHITE );

  allocate(directory.q_nrows(),directory.q_ncols(),8);

  {
    char * descr = directory.read_str(TIFFTAG_IMAGEDESCRIPTION);
    char * doc_name = directory.read_str(TIFFTAG_DOCUMENTNAME);
    if( descr != 0 || doc_name != 0 )
      if( doc_name == 0 )
	name = descr;
      else if( descr == 0 )
	name = doc_name;
      else
	name = strcat(
		 strcat(
		   strcpy(new char[strlen(descr)+strlen(doc_name)+2],
			  descr),"/"),doc_name),
        delete descr, delete doc_name;
  }


  directory.assign_to(*this);

  if( photometry == PHOTOMETRIC_MINISWHITE )
    invert();

  cout << "\n" << directory.q_ncols() << "x" << directory.q_nrows()
  	   << "x" << q_depth() << " image '"
  	   << name << "' has been read" << endl;
}

