// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   Grayscale Image
 *
 *		Write an image into a file in the TIFF format
 *
 * The present TIFF writer creates a Class G TIFF file (for gray-scale
 * images), See Appendix G of the TIFF specification. 
 *
 * The program writes the following required tags
 *	ImageWidth
 *	ImageLength
 *	RowsPerStrip	(To make strips around 8K)
 *	StripOffsets
 *	StripByteCounts
 *	XResolution	(72 dpi)
 *	YResolution	(72 dpi)
 *	ResolutionUnit
 *
 * In addition, the class G TIFF files are required to have the following
 * tags with the following values
 *	SamplesPerPixel = 1
 *	BitsPerSample   = 4 or 8
 *	Compression	= 1 or 5(LZW)
 *	PhotometricInterpretation = 0 or 1
 *
 * In addition, if the image has name, it's written under
 * 	ImageDescription,
 * The title, if specified, is written as
 *	DocumentName
 * tag
 *
 * $Id: write_tiff.cc,v 2.2 1996/03/18 19:59:53 oleg Exp oleg $
 *
 ************************************************************************
 */

#include "image.h"
#include "endian_io.h"
#ifdef __GNUC__
#pragma implementation "tiff.h"
#endif
#include "tiff.h"
#include "minmax.h"
				// Write a TIFF header into the file
void TIFFHeader::write(EndianOut& file)
{
  if( magic == TIFF_BIGENDIAN )
    file.set_bigendian();
  else if( magic == TIFF_LITTLEENDIAN )
    file.set_littlendian();
  else
    _error("Unknown magic word %x",magic);
  file.write_short(magic);
  file.write_short(version);
  file.write_long(diroffset);
}


/*
 *------------------------------------------------------------------------
 *				TIFF Directory
 * Note that all items in the TIFF directory are threaded in an ascending
 * order of their tags. 
 * Note, if some directory item has count field equal 0, the item is
 * considered "dummy" and it's not written added to the TIFF directory.
 *
 */


TIFFBeingMadeDirectory::TIFFBeingMadeDirectory(void)
  : no_entries(0), first_entry(0)
{
}


				// Insert an item into the directory
				// in an ascending order of its tag
void TIFFBeingMadeDirectory::operator += (const TIFFNewDirItem::ItemRef& ref)
{
	// should've been 
	// const_cast<TIFFNewDirItem::ItemRef&>(ref).surrender();
	// but only few compilers understand const_cast<>
  TIFFNewDirItem& new_item = 
  		(*((TIFFNewDirItem::ItemRef*)&ref)).surrender();
  no_entries++;
  if( first_entry == 0 || new_item.tag < first_entry->tag )
  {				// Add to the front of the list
    new_item.next = first_entry;
    first_entry = &new_item;
    return;
  }
  register TIFFNewDirItem * ce;
  for(ce = first_entry; ce->next != 0; ce = ce->next)
    if( new_item.tag < ce->next->tag )
      break;
  assert( new_item.tag != ce->tag );	// No duplicates allowed
  new_item.next = ce->next;
  ce->next = &new_item;
}

				// Write the entire directory
void TIFFBeingMadeDirectory::write(EndianOut& file)
{
  header.write(file);

  assert( no_entries > 0 && first_entry != 0 );
  file.write_short(no_entries);
  register TIFFNewDirItem * ce;
  for(ce=first_entry; ce != 0; ce=ce->next)
    ce->write(file);
  file.write_long(0);			// Specify this is the end of the IFD
					// no other IFD is to follow
  for(ce=first_entry; ce != 0; ce=ce->next) // Handle items that store their
  {						// values  after the directory
    const streampos pos = file.tellp();
    if( pos & 1 )			// make sure that the new item starts
      file.write_byte(0);		// at a word boundary
    ce->write_value(file);
  }
}

				// Clean up the mess: dispose of DirEntries
				// (and the entire list)
TIFFBeingMadeDirectory::~TIFFBeingMadeDirectory(void)
{
  register TIFFNewDirItem * ce=first_entry;
  while( ce != 0 )
  {
    register TIFFNewDirItem * ne = ce->next;
    delete ce;
    ce = ne;
  }
  no_entries = 0;
}

/*
 *------------------------------------------------------------------------
 *		Writing particular types of DirectoryItems
 */

				// Write this item into a file
void ScalarTIFFDE::write(EndianOut& file)
{
  file.write_short(tag);
  file.write_short(type);
  file.write_long(count);
  file.write_long(val_offset);
}

				// Write this item into a file
void StringTIFFDE::write(EndianOut& file)
{
  assert( count > 0 );
  file.write_short(tag);
  file.write_short(type);
  file.write_long(count);
  offset_for_val = file.tellp();
  file.write_long(0);			// write a dummy for now
}

				// Write the string's body.
				// Note, a padding may be necessary to
				// make the string end at a word boundary
void StringTIFFDE::write_value(EndianOut& file)
{
  streamoff pos = file.tellp();		// where are we now
					// Write a body of the string, '\0'
					// and (possible) an extra byte
  file.write((char *)val_offset,(strlen((char *)val_offset)+1+1) & ~1);
						// to keep word-boundary align
  streampos pos_new = file.tellp();
  assert( file.seekp(offset_for_val).good() );
  file.write_long(pos);			// Write the offset within the dir elem
  assert( file.seekp(pos_new).good() );	// Restore the file position 
					// to continue writing
}

				// Constructor - allocate the array
ArrayTIFFDE::ArrayTIFFDE(const short _tag, const long no_elems) 
    : TIFFNewDirItem(_tag, LONG, no_elems, 0)
{
  assert( count > 0 );
  array = new long[count];
}


				// Dispose of the array
ArrayTIFFDE::~ArrayTIFFDE(void)
{
  assert( array != 0 );
  delete [] array;
  array = 0;
}


				// Write this item into a file
				// Don't hasten to write array[0]
				// though (even if count=1): array may not
				// be filled in yet...
void ArrayTIFFDE::write(EndianOut& file)
{
  assert( count > 0 );
  file.write_short(tag);
  file.write_short(type);
  file.write_long(count);
  offset_for_val = file.tellp(),	// note the location, and
  file.write_long(0);			// write a dummy for now
}

				// Write array elements (if more than 1)
				// after the directory
void ArrayTIFFDE::write_value(EndianOut& file)
{
  long int pos;                         // Where the array is to be written
  if( count == 1 )                      // If there is only one elem, keep
    pos = array[0];                     // the offset with the dir elem itself
  else
  {
    pos = file.tellp();                 // Write the array separately
    for(register card i=0; i<count; i++)
      file.write_long(array[i]);
  }
  const streampos pos_new = file.tellp();
  assert( offset_for_val != 0 && file.seekp(offset_for_val).good() );
  file.write_long(pos);                 // Write the offset within the dir elem
  assert( file.seekp(pos_new).good() ); // Restore the file position
                                        // to continue writing
}


				// Write this item into a file: only one
				// modification: a rational has count 1, though
				// it's an array of 2 longs
void RationalTIFFDE::write(EndianOut& file)
{
  assert( count == 2 );
  file.write_short(tag);
  file.write_short(type);
  file.write_long(1);			// 2 longs, but a _single_ RATIONAL
  offset_for_val = file.tellp(),	// note the location, and
  file.write_long(0);			// write a dummy for now
}

				// Write the clone into a file
void CloneTIFFDE::write(EndianOut& file)
{
  assert( count > 0 );
  file.write_short(tag);
  file.write_short(type);
  file.write_long(count);
  if( count == 1 &&
      (type != RATIONAL && type != SRATIONAL && type != DOUBLE) )
    offset_for_val = 0,
    file.write_long(val_offset);	// write the immediate value
  else
    offset_for_val = file.tellp(),
    file.write_long(0);			// write a dummy for now
}

				// Clone the item body (from item_file)
				// unless it was small and was written
				// with the IFD itself
void CloneTIFFDE::write_value(EndianOut& file)
{
  if( offset_for_val == 0 )
    return;				// the value was written with IFD
  const streampos pos = file.tellp();
  assert( item_file.seekg(val_offset).good() );
  register card i;
  switch(type)
  {
    case LONG:
    case SLONG:
    case FLOAT:
         for(i=0; i<count; i++)
            file.write_long(item_file.read_long("reading late item"));
         break;

    case SHORT:
    case SSHORT:
         for(i=0; i<count; i++)
            file.write_short(item_file.read_short("reading late item"));
         break;
 
    case BYTE:
    case SBYTE:
    case ASCII:
    case UNDEFINED:
         for(i=0; i<count; i++)
            file.write_byte(item_file.read_byte("reading late item"));
         break;

    case RATIONAL:
    case SRATIONAL:
    case DOUBLE:
  	  for(i=0; i<count; i++)	// Copy 8-byte chunks
            file.write_long(item_file.read_long("reading late item")),
            file.write_long(item_file.read_long("reading late item"));
         break;
         
    default:
         _error("don't know how to handle type %d",type);
  }
  const streampos pos_new = file.tellp();
  assert( file.seekp(offset_for_val).good() );
  file.write_long(pos);			// Write the offset within the dir elem
  assert( file.seekp(pos_new).good() );	// Restore the file position 
					// to continue writing
}

/*
 *------------------------------------------------------------------------
 *	   	Writing strips of image's pixel matrix
 * Shred the pixel matrix into strips (containing several complete image
 * rows) of the same size (excluding possibly the last strip) and write
 * them into the file. No compression is used at present.
 * The number of rows per strip is selected such way so that
 * the strip size is around 8K (as recommended by the Standard) 
 */

class StripsInfo
{
  enum { strip_target_size = 8*1024 };
  card rows_per_strip;
  card no_strips;
  const IMAGE& image;
  unsigned long strip_size, image_size;
public:
  struct strip_info { int size; streampos offset; };
private:
  strip_info * strip_refs;
  bool written;
public:
  StripsInfo(const IMAGE& _image);
  ~StripsInfo(void)	{ delete [] strip_refs; }
  void write(EndianOut& file);
  card q_no_strips(void) const		{ return no_strips; }
  card q_rows_per_strip(void) const	{ return rows_per_strip; }
  const strip_info operator[] (const card i) const
	{ assert( written && i < no_strips ); return strip_refs[i]; }
};

StripsInfo::StripsInfo(const IMAGE& _image)
  : image(_image), written(false)
{
 rows_per_strip = strip_target_size / image.q_ncols();
 no_strips = (image.q_nrows() + rows_per_strip - 1)/rows_per_strip;
 strip_refs = new strip_info[no_strips];
 image_size = image.q_nrows() * image.q_ncols();
 strip_size = rows_per_strip * image.q_ncols();
}

			// Write strips into a file and note their locations
			// and sizes
void StripsInfo::write(EndianOut& file)
{
  if( written )
    return;

  Image_istream from_image(image);
  register unsigned long curr_im_pos;
  register strip_info * sp;
  for(sp=strip_refs,curr_im_pos=0; sp<strip_refs+no_strips; sp++)
  {
    sp->offset = file.tellp(); // where are we now, beg of a strip
    const unsigned long curr_pos_end = min(curr_im_pos+strip_size,image_size);
    sp->size = curr_pos_end - curr_im_pos;
    for(; curr_im_pos < curr_pos_end; curr_im_pos++)	// Writing a strip
       file.write_byte(from_image.get());
  }
  assert( from_image.eof() );

  written = true;
}
				// DirItem: array of strip offsets
class StripOffTIFFDE : public ArrayTIFFDE
{
protected:
  void write_value(EndianOut& file);	// Write the array of strip ptrs
					// and strips themselves
					// (after the directory is written)

  StripsInfo&	strips;

  StripOffTIFFDE(StripsInfo& _strips)
    : ArrayTIFFDE(TIFFTAG_STRIPOFFSETS,_strips.q_no_strips()),
      strips(_strips) {}
public:
  static ItemRef New(StripsInfo& strips)
  	{ return ItemRef(new StripOffTIFFDE(strips)); }
};


void StripOffTIFFDE::write_value(EndianOut& file)
{
  assert( count > 0 );
  strips.write(file);
  for(register card i=0; i<count; i++)
    array[i] = strips[i].offset;
  ArrayTIFFDE::write_value(file);
}


				// DirItem: Array of strip counts
				// It is not necessary unless compression
				// is used; but nevertheless....
class StripCntTIFFDE : public ArrayTIFFDE
{
protected:
  void write_value(EndianOut& file);	// Write the array of strip counts
					// and strips themselves
					// (after the directory is written)

  StripsInfo&	strips;

  StripCntTIFFDE(StripsInfo& _strips)
    : ArrayTIFFDE(TIFFTAG_STRIPBYTECOUNTS,_strips.q_no_strips()),
      strips(_strips) {}
public:
  static ItemRef New(StripsInfo& strips)
  	{ return ItemRef(new StripCntTIFFDE(strips)); }
};

void StripCntTIFFDE::write_value(EndianOut& file)
{
  assert( count > 0 );
  strips.write(file);
  for(register card i=0; i<count; i++)
    array[i] = strips[i].size;
  ArrayTIFFDE::write_value(file);
}

/*
 *------------------------------------------------------------------------
 *				Root module
 */

void IMAGE::write_tiff(const char * file_name,const char * title,
		       const TIFFUserAction& user_adding_tags) const
{
  is_valid();

  message("\nPreparing a TIFF file with name '%s'\n",file_name);

  EndianOut file(file_name);

  TIFFBeingMadeDirectory directory;

  directory += ScalarTIFFDE::New(TIFFTAG_IMAGEWIDTH,(unsigned)q_ncols());
  directory += ScalarTIFFDE::New(TIFFTAG_IMAGELENGTH,(unsigned)q_nrows());
  directory += ScalarTIFFDE::New(TIFFTAG_COMPRESSION,(unsigned short)COMPRESSION_NONE);
  directory += ScalarTIFFDE::New(TIFFTAG_PHOTOMETRIC,
			(unsigned short)PHOTOMETRIC_MINISBLACK);
  directory += ScalarTIFFDE::New(TIFFTAG_RESOLUTIONUNIT,
			   (unsigned short)RESUNIT_INCH);
					// Means 72 pixels per unit (inch)
  directory += RationalTIFFDE::New(TIFFTAG_XRESOLUTION,72,1);
  directory += RationalTIFFDE::New(TIFFTAG_YRESOLUTION,72,1);

  directory += ScalarTIFFDE::New(TIFFTAG_SAMPLESPERPIXEL,(unsigned short)1);
  assert( bits_per_pixel == 8 );
  directory += ScalarTIFFDE::New(TIFFTAG_BITSPERSAMPLE,(unsigned short)8);

  if( name != 0 && name[0] != '\0' ) 
    directory += StringTIFFDE::New(TIFFTAG_IMAGEDESCRIPTION, name );
  if( title != 0 && title[0] != '\0' ) 
    directory += StringTIFFDE::New(TIFFTAG_DOCUMENTNAME, title );

  StripsInfo strips_info(*this);
  directory += ScalarTIFFDE::New(TIFFTAG_ROWSPERSTRIP,
				 strips_info.q_rows_per_strip());
  directory += StripOffTIFFDE::New(strips_info);
  directory += StripCntTIFFDE::New(strips_info);

  user_adding_tags(directory);		// Give the user a chance to add
					// his own tags
  directory.write(file);			
  file.close();
}

					// The same as above, but without
					// any TIFFUserAction
void IMAGE::write_tiff(const char * file_name,const char * title) const
{
  write_tiff(file_name,title,TIFFUserAction());
}

